// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */

#include <octave/oct.h>
#include <octave/pager.h>

DEFUN_DLD (bsplinefit, args, ,
  "[a_n dx] = bsplinefit(xdata, ydata)\n"
  "\n"
  "Obtain B-spline fit coefficients of data."
  "\n"
  "Copyright 1998-2005 Spencer Olson:\n"
  "\n"
  "  xdata :  I think this is supposed to be a vector\n"
  "           xdata corresponds to the x-position of\n"
  "           of the respective data values held in ydata\n"
  "  ydata :  same dimensions as xdata\n"
)
{

    if (args.length() != 2) {
        print_usage("bsplinefit");
        return octave_value();
    }

    ColumnVector xdata( args(0).vector_value () );
    ColumnVector ydata( args(1).vector_value () );

    /* set up the grid and the data to fit structures
     * The + 2 is to accomodate the two extra endpoints (2nd derivs).
    */
    int N = xdata.length() + 2;
    double xmin = xdata(0);
    double xmax = xdata( xdata.length() - 1 );
    double dx = ( xmax - xmin ) / (N-3.0);
    double coeff = 0.0;
    int i = 0;

    ColumnVector data2fit( ydata.length() + 2 );
    data2fit(0) = ( ydata(1) - ydata(0) ) / dx;
    data2fit.insert( ydata, 1);
    data2fit( ydata.length() + 1 ) = ( ydata( ydata.length() - 1 )
                                     - ydata( ydata.length() - 2 ) ) / dx;

    Matrix B ( N, 3 );// We'll set up a tri-diagnol matrix

    for( int j=1; j < (N-1); j++ ) {
       B(j,0) = 1.0/6.0;
       B(j,1) = 2.0/3.0;
       B(j,2) = 1.0/6.0;
    }//for

    B(0,1)   = -1.0/(2.0*dx);
    B(0,2)   = 0.0;
    B(N-1,0) = 0.0;
    B(N-1,1) = 1.0/(2.0*dx);

    /* These are two special values which actually belong
     * to the (0,2) and (N-1,N-3) elements of the corresponding
     * NxN matrix.
    */
    B(0,0)   = 1.0/(2.0*dx);
    B(N-1,2) = -1.0/(2.0*dx);
    /* Now we are going to solve this using gaussian elimination for
     * tridiagnol matrices (augmented a little bit because of the
     * (0,2) and (N-1,N-3) elements alluded to earlier.
    */
    // Forward elimination %%%%
    /* First take care of the odd-ball part for the (0,2) element
     * as stored in B(0,0)
    */
    {
        coeff = B(1,0) / B(0,1);
        B(1,1) = B(1,1) - coeff*B(0,2);
        B(1,2) = B(1,2) - coeff*B(0,0);// this is the special line
        data2fit(1) = data2fit(1) - coeff*data2fit(0);
    }
    /* now we do everything up until the next odd-ball */
    for(i=2; i< (N-2); i++ ) {
        coeff = B(i-1,0) / B(i-1,1);
        B(i,1) = B(i,1) - coeff*B(i-1,2);
        data2fit(i) = data2fit(i) - coeff*data2fit(i-1);
    }//for

    /* Now we need to take care of the (N-1,N-3) element
     * stored in B(N-1,2).
    */
    {
        coeff = B(N-1,2) / B(N-3,1);
        B(N-1,0) = B(N-1,0) - coeff*B(N-2,2);// this is the special line
        data2fit(N-1) = data2fit(N-1) - coeff*data2fit(N-3);

        // Now for the very last elimination 
        coeff = B(N-1,0) / B(N-2,1);
        B(i,1) = B(i,1) - coeff*B(i-1,2);
        data2fit(i) = data2fit(i) - coeff*data2fit(i-1);
    }

    // Back Substitution %%%%
    ColumnVector a_n(N);
    a_n(N-1) = data2fit(N-1) / B(N-1,1);
    for(i=N-1; i>0; i--) {
        a_n(i) = ( data2fit(i) - B(i,2) * a_n(i+1) ) / B(i,1);
    }//for
    // And do the last substitution special because of the (0,2) element
    a_n(0) = ( data2fit(0) - B(0,2) * a_n(0+1) ) / B(0,1);

    octave_value_list retval;
    retval(0) = octave_value(a_n);
    retval(1) = octave_value(dx);
    return retval;
}

